import os

from batou.component import Attribute, Component
from batou.lib.file import File


class KeyPair(Component):

    namevar = "name"
    _required_params_ = {
        "crt": "certificate",
        "key": "key",
    }

    crt = None
    key = None

    base_path = Attribute(str, "")
    provide_itself = Attribute(bool, default=True)

    def configure(self):
        self.crt_file = File(
            os.path.join(self.base_path, "{}.crt".format(self.name)),
            content=self.crt,
        )
        self += self.crt_file
        self.key_file = File(
            os.path.join(self.base_path, "{}.key".format(self.name)),
            content=self.key,
            mode=0o600,
        )
        self += self.key_file

        if self.provide_itself:
            self.provide("keypair::{}".format(self.name), self)



#!/usr/bin/env python

import logging
from argparse import ArgumentParser
from dataclasses import dataclass
from logging import info, warning
from math import ceil
from os import environ, execvp, fork
from subprocess import check_call
from time import sleep

import requests


@dataclass
class HealthCheckResult:
    success: bool
    summary: Exception | None = None

    def __bool__(self) -> bool:
        return self.success

    def __str__(self) -> str:
        assert self.summary is not None
        return str(self.summary)


# Sentinel value for a successful healthcheck. Doesn't
# really make sense to instantiate a new HealthCheckResult whenever
# a healthcheck passes.
HEALTHCHECK_SUCCESS = HealthCheckResult(True)


class ExponentialBackoff:
    def __init__(self, static_check_interval: int):
        self.static_check_interval = static_check_interval
        self.last_healthcheck = HEALTHCHECK_SUCCESS
        self.__reset()

    def failure(self, last_failure_reason: HealthCheckResult):
        self.n_failures += 1
        self.last_healthcheck = last_failure_reason

    def sleep(self, timeout: int):
        seconds_to_sleep = self.static_check_interval + (2**self.n_failures)
        if self.last_healthcheck:
            self.__reset()
        else:
            self.__report_failing_healthcheck(
                seconds_to_sleep,
                self.__will_sleep_exceed_watchdog(timeout, seconds_to_sleep),
            )
            self.last_healthcheck = HEALTHCHECK_SUCCESS

        sleep(seconds_to_sleep)

    def __report_failing_healthcheck(
        self, sleep_seconds: int, will_time_out: bool
    ):
        log_message = f"Healthcheck failure (Reason: {self.last_healthcheck}), sleeping {sleep_seconds}."
        if will_time_out:
            warning(f"{log_message} Watchdog will likely kill process")
        else:
            info(log_message)

    def __will_sleep_exceed_watchdog(
        self, watchdog_timeout: int, seconds_to_sleep: int
    ) -> bool:
        # Seconds that were slept so far + seconds that will be slept.
        # This doesn't take the time spent with the healthcheck into account,
        # but that's only an approximation for logging.
        total_sleep = (
            # Seconds to be slept on the next sleep
            seconds_to_sleep
            # 2^n seconds of sleep for each 0 < n < self.n_failures
            + (2**self.n_failures - 2)
            # on each sleep it was additionally slept check_interval seconds
            + (self.n_failures - 1) * self.static_check_interval
        )

        return total_sleep >= watchdog_timeout

    def __reset(self):
        self.n_failures = 0


def is_service_available(url: str, timeout: int) -> HealthCheckResult:
    try:
        requests.get(url, timeout=timeout).raise_for_status()
        return HEALTHCHECK_SUCCESS
    except Exception as ex:
        return HealthCheckResult(False, ex)


def await_service(
    url: str, healthcheck_timeout: int, startup_loop_interval: int
):
    # No need to handle timeouts here: due to `Type=notify`,
    # this unit won't be up until this loop has terminated.
    # The timeout for that can be controlled in the unit directly
    # via TimeoutStartSec from `systemd.service(5)`.
    while not is_service_available(url, healthcheck_timeout):
        info(f"Service isn't up yet, sleeping {startup_loop_interval}s")
        sleep(startup_loop_interval)

    # Tell the service-manager that we're ready.
    # Because of this, the unit's state transitions from 'activating' to 'active'.
    check_call(["systemd-notify", "--ready"])
    info("Service ready, now monitoring.")


def monitor_service(
    url: str,
    watchdog_timeout: int,
    healthcheck_timeout: int,
    min_check_interval: int,
):
    # For WatchdogSec=32 with a sleep of 1s between all attempts there are three attempts
    # to recover.
    # I.e. 2^n seconds of sleep + 1s of sleep interval.
    # For each of the three attempts so far this is
    #   2^1 + 1 + 2^2 + 1 * 2^3 + 1 = 17s
    # because this is the fourth failure, a sleep of 2^4 + 1 seconds (=17s) will
    # be started . Since 17+17 > 32, the watchdog will kill the process before another
    # attempt can be made.
    #
    # The extra 1s (=min_check_interval) is added to provide a configurable grace period
    # between healthchecks. In some cases it may not be desirable to issue a healthcheck every second.
    exp_backoff = ExponentialBackoff(min_check_interval)

    while True:
        if result := is_service_available(url, healthcheck_timeout):
            check_call(["systemd-notify", "WATCHDOG=1"])
        else:
            exp_backoff.failure(result)

        exp_backoff.sleep(watchdog_timeout)


if __name__ == "__main__":
    argparser = ArgumentParser()
    argparser.add_argument(
        "--healthcheck-url",
        help="URL to issue a request against to check the service's health",
    )
    argparser.add_argument(
        "--healthcheck-timeout",
        default=2,
        type=int,
        help="How many seconds until the healthcheck request times out",
    )
    argparser.add_argument(
        "--watcher-loglevel",
        default="warning",
        help="Loglevel of the watchdog script, doen't influence the main process",
    )
    argparser.add_argument(
        "--startup-check-interval",
        default=4,
        type=int,
        help="Seconds to wait between healthchecks when the service is starting up",
    )
    argparser.add_argument(
        "--check-interval",
        default=2,
        type=int,
        help="Seconds to wait between healtchecks when the service is running",
    )
    argparser.add_argument(
        "command",
        nargs="+",
        help="Service process that will be watched by this script",
    )
    args = argparser.parse_args()

    # After how much time (in microseconds) the watchdog must be pet.
    # Set by systemd directly.
    assert (
        "WATCHDOG_USEC" in environ
    ), "WATCHDOG_USEC not in environment, please configure WatchdogSec in the systemd service!"
    watchdog_timeout = int(environ["WATCHDOG_USEC"]) // 1_000_000

    pid = fork()
    if pid > 0:
        # By making the service the main process, it will be restarted immediately when
        # the service crashes rather than waiting for the exponential backoff to fail.
        pass_args = args.command
        execvp(pass_args[0], pass_args)
    else:
        logging.basicConfig(level=args.watcher_loglevel.upper())
        info("Starting watcher for application as child process")

        await_service(
            args.healthcheck_url,
            args.healthcheck_timeout,
            args.startup_check_interval,
        )
        monitor_service(
            args.healthcheck_url,
            watchdog_timeout,
            args.healthcheck_timeout,
            args.check_interval,
        )



import batou.component
import batou.lib.file


class Run(batou.component.Component):
    """
    A component taking over a script content, saving it to an actual script
    and running it afterwards.
    Example:

    script_content = '''
    #!/bin/sh
    echo "hello world"
    '''
    self += Run(
        'hello.sh',
        content=script_content)

    You can also provide ready-to-use-script by yourself:

    self += batou.lib.file.File('myscript', mode=0o744)
    self += batou_ext.run.Run('myscript', file=self._)

    The script is called only, if there are any changes at the parent
    component.
    """

    _required_params_ = {
        "content": "run!",
    }
    namevar = "command"
    content = None
    file = None
    env = None

    def configure(self):

        if not self.file and self.content:
            self += batou.lib.file.File(
                self.command, content=self.content, mode=0o700
            )
            self.command_file = self._
        elif self.file:
            self.command_file = self.file
        else:
            raise ValueError("You need to provide either content or file.")

    def verify(self):
        self.parent.assert_no_changes()
        self.assert_file_is_current(
            f"{self.command_file.path}_stamp",
            requirements=[
                f"{self.command_file.path}",
            ],
        )

    def update(self):
        self.touch(self.command_file.path)
        self.cmd(self.command_file.path, env=self.env)
        self.touch(f"{self.command_file.path}_stamp")



import difflib

import batou
from batou.component import Attribute, Component


class ACL(Component):
    """Replace a file's ACL."""

    path = ""
    namevar = "path"

    # The complete new ACL. See `getfacl <filename>` for the format.
    ruleset = Attribute(list)  # of str

    def update(self):
        rules = "\n".join(self.ruleset).encode("UTF-8")
        proc = self.cmd(
            f"setfacl --set-file=- '{self.path}'",
            communicate=False,
        )
        outs, errs = proc.communicate(input=rules)
        if proc.returncode > 0:
            raise ValueError(f"setfacl error: {errs}")

    def verify(self):
        proc = self.cmd(f"getfacl -cpE '{self.path}'", communicate=False)
        outs, errs = proc.communicate()
        outs = outs.decode("UTF-8", errors="replace").strip()
        current_rules = sorted(outs.strip().split("\n"))
        if current_rules != sorted(self.ruleset):
            d = difflib.Differ()
            l1 = "\n".join(current_rules).splitlines()
            l2 = "\n".join(sorted(self.ruleset)).splitlines()
            result = d.compare(l1, l2)
            self.log("Updating ACL, diff:\n" + "\n".join(result))

            raise batou.UpdateNeeded



import glob
import os
import os.path
import shutil
import subprocess
import tempfile
import urllib.parse
from textwrap import dedent

import batou
import batou.component
import batou.lib.file
import batou.lib.git

import batou_ext.ssh


class SymlinkAndCleanup(batou.component.Component):
    """
    Symlink the give file or directory to `current`, symlink the file or directory that was symlinked to `current` to `last`
    and remove all files matching `pattern` that are not currently linked to current or last.
    This utility is mostly used for Git Checkouts, which can grow considerably in size, but may also be used for
    S3 File Downloads or other Folders and Files that should be cached.

    The files matching `pattern` are discarded of using the `DeploymentTrash` component.
    Under the hood it uses systemd's tmpfiles.

    ```python
    new_file = download("test-v1.0.tar.gz")
    # clean up a all downloaded versions but the last two
    self += SymlinkAndCleanup(new_file.path, pattern = "*.tar.gz")
    ```
    """

    namevar = "current"
    pattern = None
    _current_link = None
    _last_link = None

    prefix = None

    systemd_read_max_iops = 100
    systemd_write_max_iops = 100
    trashdir = None

    ## DEPRECATED, do not use
    use_systemd_run_async_cleanup = False
    systemd_extra_args = None

    def configure(self):
        self._current_link = (
            f"{self.prefix}-current" if self.prefix else "current"
        )
        self._last_link = f"{self.prefix}-last" if self.prefix else "last"
        self.dir = os.path.dirname(self.current)
        self.current = os.path.basename(self.current)
        self += DeploymentTrash(
            read_iops_limit=self.systemd_read_max_iops,
            write_iops_limit=self.systemd_write_max_iops,
            trashdir=self.trashdir,
        )
        self.trash = self._

        if self.use_systemd_run_async_cleanup:
            batou.output.annotate(
                "use_systemd_run_async_cleanup is deprecated and will be removed in a future release, please remove it",
                yellow=True,
            )

        if self.systemd_extra_args:
            batou.output.annotate(
                "systemd_extra_args is deprecated and will be removed in a future release, please remove it",
                yellow=True,
            )

    @staticmethod
    def _link(path):
        try:
            return os.readlink(path)
        except OSError:
            batou.output.error(f"Unable to read link: `{path}`")
            return None

    @staticmethod
    def remove(links, el):
        try:
            links.remove(el)
        except ValueError:
            pass

    def _list_removals(self):
        candidates = glob.glob(self.pattern)

        current = self._link(self._current_link)
        last = self._link(self._last_link)
        if current == self.current:
            # keep last+current
            self.remove(candidates, current)
            self.remove(candidates, last)
        else:
            # keep current + new current"
            self.remove(candidates, current)
            self.remove(candidates, self.current)

        return candidates

    def verify(self):
        with self.chdir(self.dir):
            if self._link(self._current_link) != self.current:
                raise batou.UpdateNeeded()
            if self._list_removals():
                raise batou.UpdateNeeded()

    def update(self):
        with self.chdir(self.dir):
            current = self._link(self._current_link)
            if current != self.current:
                try:
                    os.remove(self._current_link)
                except OSError as e:
                    batou.output.error(
                        f'Failed to remove link "{self._current_link}" ({current}): {e.strerror}'
                    )
                    pass

                try:
                    os.remove(self._last_link)
                except OSError as e:
                    last = self._link(self._last_link)
                    batou.output.error(
                        f'Failed to remove link "{self._last_link}" ({last}): {e.strerror}'
                    )
                    pass

                batou.output.annotate("current -> {}".format(self.current))
                os.symlink(self.current, self._current_link)
                if current:
                    batou.output.annotate("last -> {}".format(current))
                    os.symlink(current, self._last_link)

            candidates = self._list_removals()
            if not candidates:
                batou.output.annotate("Nothing to remove.")

            for c in candidates:
                batou.output.annotate(f"Removing: {candidates}")
                self.trash.discard(c)


class DeploymentTrash(batou.component.Component):
    """
    A trash folder that is regularly cleaned up by systemd's tmpfiles.
    Files and Folders can be moved in here for asynchronous deletion that is not tied to the deployment.
    Due to this component being commonly used to offload deleting very large files, it also supports adding IOPS limits.


    ```python
    self.trash = DeploymentTrash(read_iops_limit = 250, write_iops_limit = 250)
    self += self.trash

    # files that are created here should not be part of the deployment but rather a sideproduct
    # for example a cache directory or build artefacts that are generated by some script
    large_folder = do_something_that_generates_a_lot_of_byproducts()
    self.trash.discard(large_folder)
    ```
    """

    read_iops_limit = 100
    write_iops_limit = 100

    trashdir = None

    def configure(self):
        if not self.trashdir:
            self.trashdir = os.path.expanduser("~/.deployment-trash")

        self += batou.lib.file.File(self.trashdir, ensure="directory")
        self += batou.lib.file.File(
            "/etc/local/nixos/trash.nix",
            content=dedent(
                """\
            {
              systemd.tmpfiles.rules = [
                "d {{component.trashdir}} 0755 - - 1h -"
              ];

              systemd.services."systemd-tmpfiles-clean".serviceConfig = {
                IOReadIOPSMax="{{component.trashdir}} {{component.read_iops_limit}}";
                IOWriteIOPSMax="{{component.trashdir}} {{component.write_iops_limit}}";
              };
            }
        """
            ),
        )

    def discard(self, path):
        target = tempfile.mkdtemp(dir=self.trashdir)
        try:
            os.rename(path, os.path.join(target, os.path.basename(path)))
        except FileNotFoundError:
            # Nothing to delete.
            pass
        except OSError as e:
            batou.output.annotate(
                "check that the deployment trash dir and the directory you want to trash are on the same device"
            )
            raise e



import collections
import hashlib
import inspect
import json
import os
import os.path
import shlex
import subprocess
import time
from pathlib import Path

import batou
import batou.component
import batou.lib.cron
import batou.lib.file
import batou.lib.logrotate
import batou.lib.nagios
import batou.lib.service
import batou.lib.supervisor
import batou.utils
import pkg_resources
from batou import (
    IPAddressConfigurationError,
    ReportingException,
    UpdateNeeded,
    output,
)
from batou.component import Component, RootComponent
from batou.environment import Environment
from batou.host import Host
from batou.lib.file import (
    File,
    Group,
    ManagedContentBase,
    Mode,
    Owner,
    Presence,
)
from batou.utils import Address, NetLoc


class Package(batou.component.Component):
    """Install Nix package for user.

    Usage:

        self += batou_ext.nix.Package(attribute='nixos.yarn')
        self += batou_ext.nix.Package('pbzip2-1.1.12')

    """

    package = None
    attribute = None
    file = None

    def __init__(self, namevar=None, **kw):
        # Make 'namevar' optional
        kw["package"] = namevar
        super(Package, self).__init__(**kw)

    def configure(self):
        if self.file:
            assert self.package
            if not os.path.isabs(self.file):
                self.file = os.path.join(self.defdir, self.file)

    def verify(self):
        if self.attribute:
            stdout, stderr = self.cmd("nix-env -qaA {{component.attribute}}")
            self._installs_package = stdout.strip()
        else:
            self._installs_package = self.package
        stdout, stderr = self.cmd("nix-env --query")
        if self._installs_package not in stdout.splitlines():
            raise batou.UpdateNeeded()

    def update(self):
        if self.attribute:
            self.cmd("nix-env -iA {{component.attribute}}")
        elif self.file:
            self.cmd("nix-env -if {{component.file}}")
        else:
            self.cmd("nix-env -i {{component.package}}")

    @property
    def namevar_for_breadcrumb(self):
        return self.attribute if self.attribute else self.package


class PurgePackage(batou.component.Component):
    namevar = "package"

    def verify(self):
        try:
            self.cmd(f"nix-env --query {self.package}")
            raise batou.UpdateNeeded()
        except batou.utils.CmdExecutionError as e:
            if e.stderr.strip().endswith("matches no derivations"):
                batou.output.annotate(
                    f"Could not find package to purge: {self.package}",
                    yellow=True,
                )
            else:
                e.report()

    def update(self):
        self.cmd("nix-env --uninstall {{component.package}}")


class UserEnv(batou.component.Component):
    """Provide a NixOS user environment.

    Usage::
        self += batou_ext.nix.UserEnv(
                "django",
                packages=[
                    "gcc",
                    "gettext",
                    "glibc",
                    "liberation_ttf",
                    "libffi",
                    "libxml2",
                    "libxslt",
                    "mysql55",
                    "nodejs-8_x",
                    "openssl",
                    "python27Full",
                    "yarn",
                    "zip",
                ],
                channel="<nixos-channel-base-url-to-be-put-in>/nixexprs.tar.xz"
                shellInit="# additional shell init")

    A list of available channels can be found at e.g.
    https://nixos.org/channels/
    """

    namevar = "profile_name"
    channel = batou.component.Attribute(str)
    shellInit = ""
    packages = ()
    let_extra = ""

    def configure(self):
        self.checksum = hashlib.sha256()
        template = pkg_resources.resource_string(
            __name__, "resources/userenv.nix"
        ).decode("UTF-8")
        self.profile = self.expand(template).encode("UTF-8")
        self.checksum.update(self.profile)
        self.nix_env_name = self.expand(
            "{{component.profile_name}}-1.{{component.checksum.hexdigest()}}"
        )
        self += batou.lib.file.File(
            self.profile_name + ".nix", content=self.profile
        )
        self += Package(self.nix_env_name, file=self._.path)
        self.user_profile_path = os.path.expanduser(
            "~/.nix-profile/etc/profile.d/{}.sh".format(self.profile_name)
        )
        self.user_bin_path = os.path.expanduser("~/.nix-profile/bin")


class Rebuild(batou.component.Component):
    """Trigger rebuild on FC platform.

    Usage::

        # Tirgger rebuild if self or subcomponent changed
        self += batou_ext.nix.Rebuild()

        # Trigger rebuild if specific components changed:
        self += batou_ext.nix.Rebuild(dependencies=(self, foo, bar))

    """

    dependencies = None

    def verify(self):
        if self.dependencies:
            for dependency in self.dependencies:
                dependency.assert_no_changes()
        else:
            self.parent.assert_no_subcomponent_changes()

    def update(self):
        self.cmd("sudo fc-manage --build")


def rebuild(cls):
    """Class decctorator easily allow rebuild in multi platform environments.

    This is possible and useful because restarting or reconfiguring services
    on NixOS always is *just* a rebuild.

    Usage::

        @batou_ext.nix.rebuild
        class Foo(batou.component.Component):
            ...

    """
    cls._add_platform("nixos", Rebuild)
    return cls


# Uhugh. Patch service so we can set checksum/systemd data which is being used
# by UserInit.
batou.lib.service.Service.checksum = ""
batou.lib.service.Service.systemd = None


@batou.component.platform("nixos", batou.lib.service.Service)
class UserInit(batou.component.Component):
    """Start services on fc platform.

    Usage:

    * Import this module at least once in your deployment.
    * Set platform in environment to `nixos`.
    """

    def configure(self):
        self.executable = self.parent.executable
        name = shlex.split(self.executable)[0]
        name = os.path.basename(name)
        self.name = os.path.splitext(name)[0]

        # Remove older versions that were created with
        # hard-to-read or broken service filenames to avoid accidental
        # duplicates.
        service_file = "/etc/local/systemd/{}.service".format(self.name)
        old_service_file = "/etc/local/systemd/{}.service".format(
            os.path.basename(self.executable)
        )
        if old_service_file != service_file:
            self += batou.lib.file.Purge(old_service_file)

        # The systemd config namespace of our expected systemd dictionary
        # is a little messed up.
        #
        # You can use
        # - bare config attribute names, those will end up in the [Service]
        #   section
        # - <Section>_<Attribute>, those will end up in the respective
        #   [<Section>] sections.
        #
        # E.g. systemd=dict(Unit_After='afdsfdasfsda', Service_Type='Simple',
        #                  User='nobody')
        config = {}
        config["Unit"] = {"X-Restart-Triggers": self.parent.checksum}
        config["Service"] = dict(
            Type="forking",
            LimitNOFILE="64000",
            LimitNPROC="64173",
            LimitSIGPENDING="64173",
            User=self.environment.service_user,
            Group="service",
            ExecStart=os.path.join(self.root.workdir, self.executable),
            Restart="always",
            Environment=[
                self.expand(
                    "LOCALE_ARCHIVE={{component.env['LOCALE_ARCHIVE']}}"
                ),
                self.expand("PATH={{component.env['PATH']}}"),
                self.expand("TZDIR={{component.env['TZDIR']}}"),
            ],
        )

        # Phase 1: allow overriding keys, do not append lists
        systemd = self.parent.systemd or {}
        for key, value in list(systemd.items()):
            if "_" in key:
                section, key = key.split("_", 1)
            else:
                section = "Service"
            config.setdefault(section, {})[key] = value

        # Phase 2: expand (trivial) iterables into multiple keys
        self.config = {}
        for section in config:
            self.config[section] = []
            for key, value in list(config[section].items()):
                if not isinstance(value, (list, tuple)):
                    value = [value]
                for v in value:
                    self.config[section].append((key, v))

        self.checksum = self.parent.checksum
        self += batou.lib.file.File(
            service_file,
            content=pkg_resources.resource_string(
                "batou_ext", "resources/systemd.service"
            ),
        )
        self += Rebuild()

    @property
    def env(self):
        env = collections.defaultdict(dict)
        env.update(os.environ)
        return env

    def start(self):
        self.cmd("sudo systemctl start {}".format(self.name))

    def verify(self):
        self.assert_cmd("sudo systemctl is-active {}".format(self.name))

    def update(self):
        self.start()


@batou.component.platform("nixos", batou.lib.supervisor.RunningSupervisor)
class FixSupervisorStartedBySystemd(batou.component.Component):
    """For some, yet unknown reason, supervisor is not started with systemd.

    To fix it, we need to restart supervisor.

    """

    def verify(self):
        if not self.parent.is_running():
            # Well, if it's not running, something is really hosed.
            return
        try:
            self.cmd("sudo systemctl status supervisord")
        except batou.utils.CmdExecutionError:
            # Aha. IT's running, but not by systemd!
            raise batou.UpdateNeeded

    def update(self):
        self.cmd("bin/supervisorctl shutdown")
        wait = 60
        while wait:
            out, err = "", ""
            out, err = self.cmd("bin/supervisorctl pid", ignore_returncode=True)
            if "no such file" in out:
                break
            time.sleep(1)
            wait -= 1

        self.cmd("sudo systemctl start supervisord")


@batou.component.platform("nixos", batou.lib.cron.CronTab)
class InstallCrontab(batou.lib.cron.InstallCrontab):
    """Install crontab with crontab command."""

    def update(self):
        self.cmd(self.expand("cat {{component.crontab.path}} | crontab -"))


# Patch Service so we can pass down the name to sensu
batou.lib.nagios.Service.name: str = None
batou.lib.nagios.Service.interval: int = 60
batou.lib.nagios.Service.cron: str = None


@rebuild
class SensuChecks(batou.component.Component):
    """SensuChecks gathers ServiceChecks and creates sensu checks accordingly.

    This is a substitute for batou.lib.nagios.NRPEHost.

    `ServiceCheck` instances require an additional attribute `name`, to
    identify the check::

        self += ServiceCheck(
            'Testname',
            name='accounting',
            ...)


    The default execution interval is 60 seconds, and can be changed by
    setting `interval` on the ServiceCheck instance, e.g.::

        self += ServiceCheck(
            'Testname',
            name='accounting',
            command='/path/to/checker',
            args="--my-check-args",
            interval=120)

    Sensu also supports a cron notation:

        self += ServiceCheck(
            'Testname',
            name='accounting',
            command='/path/to/checker',
            args="--my-check-args",
            interval=None,
            cron="*5/ 3-7 * * *")


    """

    purge_old_batou_json = batou.component.Attribute("literal", default=True)

    def configure(self):
        self.services = self.require(
            batou.lib.nagios.Service.key, host=self.host
        )
        checks = {}
        for service in self.services:
            assert getattr(service, "name", None)
            checks[service.name] = check = dict(
                standalone=True,
                command=service.expand(
                    "{{component.command}} {{component.args}}"
                ),
            )

            if service.interval:
                check["interval"] = service.interval
            elif service.cron:
                check["cron"] = service.cron
            else:
                raise ValueError("Need either `interval` or `cron` setting.")

        config_file_name = "/etc/local/sensu-client/{}-batou.json".format(
            self.environment.service_user
        )

        if checks:
            sensu = dict(checks=checks)

            self += batou.lib.file.File(
                config_file_name,
                content=json.dumps(sensu, sort_keys=True, indent=4),
            )
        else:
            self += batou.lib.file.Purge(config_file_name)

        if self.purge_old_batou_json:
            self += batou.lib.file.Purge("/etc/local/sensu-client/batou.json")


@batou.component.platform("nixos", batou.lib.logrotate.Logrotate)
class LogrotateIntegration(batou.component.Component):
    def configure(self):
        assert (
            self.environment.service_user
        ), "Need to set service_user inside environment file."
        user = self.environment.service_user
        user_logrotate_conf = os.path.join(
            "/etc/local/logrotate", user, "batou.conf"
        )
        self += batou.lib.file.File(
            user_logrotate_conf, content=self.parent.logrotate_conf.content
        )


class PythonWithNixPackages(batou.component.Component):
    """Provide a python interpreter which contains packages from nix.

    Usage:


    Provides "python3.5" pointing to the path of the executable.

    """

    _required_params_ = {"python": "python3.5"}
    # python interpreter, "python3.5"
    python = None

    # sequence of nix package attributes
    # this must include the actual python, .i.e.  pkgs.python34
    nix_packages = ()
    pythonPackages = None

    def configure(self):
        if not self.pythonPackages:
            self.pythonPackages = "pkgs.{}Packages".format(
                self.python.replace(".", "")
            )

        self += batou.lib.file.File(
            "{}.nix".format(self.python),
            content=pkg_resources.resource_string(
                "batou_ext", "resources/python.nix"
            ),
        )
        self += batou.lib.file.File(
            "setupEnv-{}".format(self.python),
            mode=0o755,
            content=pkg_resources.resource_string(
                "batou_ext", "resources/setupEnv.sh"
            ),
        )
        self.env_file = self._
        self += batou.lib.file.File(
            "{}.c".format(self.python),
            content=pkg_resources.resource_string(
                "batou_ext", "resources/loader.c"
            ),
        )
        self.provide(self.python, os.path.join(self.workdir, self.python))

    def verify(self):
        self.assert_no_subcomponent_changes()

    def update(self):
        self.cmd("gcc {}.c -o {}".format(self.python, self.python))
        # Start up once to load all dependencies here and not upon the first
        # use:
        self.cmd("./{} -c True".format(self.python))


def nix_dict_to_nix(dct):
    """Converts a dict with values that are already nixified to Nix code."""
    content = " ".join(f"{n} = {v};" for n, v in dct.items())
    return "{ " + content + " }"


def seq_to_nix(seq):
    content = " ".join(value_to_nix(v) for v in seq)
    return "[ " + content + " ]"


def mapping_to_nix(obj):
    # XXX: only str keys for now

    converted = {}
    for k, v in obj.items():
        conv = value_to_nix(v)
        if conv is not None:
            converted[k] = conv
    return nix_dict_to_nix(converted)


def str_to_nix(value):
    # https://nixos.org/manual/nix/stable/language/values.html#type-string
    value = (
        value.replace("\\", "\\\\").replace("${", "\\${").replace('"', '\\"')
    )
    return f'"{value}"'


def environment_to_nix_dict(env: Environment):
    dct = {
        "base_dir": str_to_nix(env.base_dir),
        "connect_method": str_to_nix(env.connect_method),
        "deployment_base": str_to_nix(env.deployment_base),
        "name": str_to_nix(env.name),
        "target_directory": str_to_nix(env.target_directory),
        "workdir_base": str_to_nix(env.workdir_base),
    }

    if env.host_domain is not None:
        dct["host_domain"] = str_to_nix(env.host_domain)
    if env.platform is not None:
        dct["platform"] = str_to_nix(env.platform)
    if env.service_user is not None:
        dct["service_user"] = str_to_nix(env.service_user)

    return dct


def netloc_to_nix_dict(netloc: NetLoc):
    return {
        "__toString": f'_: "{netloc}"',
        "host": str_to_nix(netloc.host),
        "port": str(netloc.port),
    }


def address_to_nix_dict(addr: Address):
    dct = {
        "__toString": f"_: {str_to_nix(str(addr))}",
        "connect": nix_dict_to_nix(netloc_to_nix_dict(addr.connect)),
    }
    try:
        dct["listen"] = nix_dict_to_nix(netloc_to_nix_dict(addr.listen))
    except IPAddressConfigurationError:
        pass
    try:
        dct["listen_v6"] = nix_dict_to_nix(netloc_to_nix_dict(addr.listen_v6))
    except IPAddressConfigurationError:
        pass

    return dct


def host_to_nix_dict(host: Host):
    return {"fqdn": str_to_nix(host.fqdn), "name": str_to_nix(host.name)}


def value_to_nix(value):
    if isinstance(value, str):
        return str_to_nix(value)
    elif isinstance(value, bool):
        return str(value).lower()
    elif value is None:
        return None
    elif isinstance(value, int):
        return str(value)
    elif isinstance(value, Path):
        return str(value)
    elif isinstance(value, dict):
        return mapping_to_nix(value)
    elif isinstance(value, list):
        return seq_to_nix(value)
    elif isinstance(value, tuple):
        return seq_to_nix(value)
    elif isinstance(value, Component):
        return component_to_nix(value)
    elif isinstance(value, Address):
        return nix_dict_to_nix(address_to_nix_dict(value))
    elif isinstance(value, Host):
        return nix_dict_to_nix(host_to_nix_dict(value))
    elif isinstance(value, Environment):
        return nix_dict_to_nix(environment_to_nix_dict(value))
    elif isinstance(value, batou.utils.Timer):
        return None  # ignore
    else:
        raise TypeError(f"unsupported type '{type(value)}'")


def component_to_nix(component: Component):
    from batou_ext.nixos import NixOSModuleContext

    attrs = {}

    for name, value in inspect.getmembers(component):
        if name.startswith("_"):
            pass
        elif value is component:
            pass
        elif inspect.ismethod(value) or inspect.isgenerator(value):
            pass
        elif name in ("sub_components", "changed", "instances"):
            pass
        elif isinstance(value, NixOSModuleContext):
            pass
        elif isinstance(value, RootComponent):
            if (
                value.component is not component
                and component.parent is not value.component
            ):
                attrs[name] = component_to_nix(value.component)
        elif isinstance(value, Component):
            if value is not component.parent:
                attrs[name] = component_to_nix(value)
        elif isinstance(value, NixOSModuleContext):
            pass
        else:
            try:
                converted_value = value_to_nix(value)
                if converted_value is not None:
                    attrs[name] = converted_value
            except TypeError as e:
                component.log(f"Cannot convert {name}: {e.args[0]}")

    return nix_dict_to_nix(attrs)


class NixSyntaxCheckFailed(ReportingException):
    def __init__(self, error_msg, path=None):
        self.error_msg = error_msg.strip().removeprefix("error: ")
        self.path = path

    def __str__(self):
        return f"Nix syntax check failed: {self.error_msg} in {self.path}"

    def report(self):
        output.error(f"Nix check {self.error_msg}")


class NixContent(ManagedContentBase):
    format_nix_code = False
    check_nix_syntax = True

    def render(self):
        pass

    def verify(self, predicting=False):
        update_needed = False

        if self.format_nix_code:
            try:
                proc = subprocess.run(
                    ["nixfmt"],
                    input=self.content,
                    check=True,
                    capture_output=True,
                )
                self.content = proc.stdout
            except FileNotFoundError:
                self.log("Cannot format Nix file, nixfmt not found.")
            except subprocess.CalledProcessError as e:
                self.log(f"nixfmt failed: {e.stderr}")

        try:
            super().verify(predicting)
        except UpdateNeeded:
            update_needed = True

        if self.check_nix_syntax:
            try:
                subprocess.run(
                    ["nix-instantiate", "--parse", "-"],
                    input=self.content,
                    check=True,
                    capture_output=True,
                )
            except FileNotFoundError:
                self.log(
                    "Cannot syntax-check Nix file, nix-instantiate not found."
                )
            except subprocess.CalledProcessError as e:
                raise NixSyntaxCheckFailed(
                    e.stderr.decode("utf8"), path=self.path
                )

        if update_needed:
            raise UpdateNeeded()


class NixFile(File):
    format_nix_code = False

    def configure(self):
        self._unmapped_path = self.path
        self.path = self.map(self.path)
        self += Presence(self.path, leading=self.leading)

        # variation: content or source explicitly given

        # The mode needs to be set early to allow batou to get out of
        # accidental "permission denied" situations.
        if self.mode:
            self += Mode(self.path, mode=self.mode)

        # no content or source given but file with same name
        # exists
        if self.content is None and not self.source:
            guess_source = self.root.defdir + "/" + os.path.basename(self.path)
            if os.path.isfile(guess_source):
                self.source = guess_source
            else:
                # Avoid the edge case where we want to support a very simple
                # case: specify File('asdf') and have an identical named file
                # in the component definition directory that will be templated
                # to the work directory.
                #
                # However, if you mis-spell the file, then you might
                # accidentally end up with an empty file in the work directory.
                # If you really want an empty File then you can either use
                # Presence(), or (recommended) use File('asdf', content='') to
                # make this explicit. We don't want to accidentally confuse the
                # convenience case (template a simple file) and an edge case
                # (have an empty file)
                raise ValueError(
                    "Missing implicit template file {}. Or did you want "
                    "to create an empty file? Then use File('{}', content='').".format(
                        guess_source, self._unmapped_path
                    )
                )
        if self.content or self.source:
            content = NixContent(
                self.path,
                source=self.source,
                encoding=self.encoding,
                content=self.content,
                sensitive_data=self.sensitive_data,
                format_nix_code=self.format_nix_code,
            )
            self += content
            self.content = content.content

        if self.owner:
            self += Owner(self.path, owner=self.owner)

        if self.group:
            self += Group(self.path, group=self.group)



import batou.component
import batou.lib.file
import pkg_resources

import batou_ext.nix


class HTTPBasicAuth(batou.component.Component):
    """A glueing component for managing http_basic_auth auth strings.

    This component _does not_ generate the htpasswd string itself, but
    requires a pre-compiled one (stored into secrets). You may use

    mkpasswd -R 100000 -m sha-512 <password> <salt>

    for generating the passwordsection. General structure of entry is like
    <username>:<generated_password_string>

    Usage::

        from batou_ext.http import HTTPBasicAuth

        self += batou_ext.http.HTTPBasicAuth(
            env_name="myauth",
            username="joe",
            password="secret",
            basic_auth_string="joe:$apr1$Ma0Fc6pW$Kl5dV4ecBXH12gDieRHVq.")

    Alternatively, if you want to use Flyingcircus' autogenerated http_auth
    file based on project configuration, just use::

        self += batou_ext.http.HTTPBasicAuth(fcio_auth=True)

    """

    _required_params_ = {"env_name": "test", "basic_auth_string": "secret"}
    env_name = None
    fcio_auth = batou.component.Attribute("literal", default=False)
    username = None
    password = None
    basic_auth_string = None
    providing = batou.component.Attribute("literal", default=True)

    def configure(self):
        if self.providing:
            self.provide("http_basic_auth", self)
        if self.fcio_auth:
            self.path = "/etc/local/nginx/htpasswd_fcio_users"
        else:
            self._deploy_customer_http_auth_file()

    def _deploy_customer_http_auth_file(self):
        if self.env_name is None:
            raise ValueError(
                "You need to define an environment name for "
                "your http_baisc_auth-file"
            )
        self += batou.lib.file.File(
            "htpasswd_{}".format(self.env_name), content=self.basic_auth_string
        )
        self.path = self._.path


@batou_ext.nix.rebuild
class HTTPServiceWatchdog(batou.component.Component):
    """
    Adds a "watchdog" on top of a systemd service that checks within a given interval
    whether an HTTP URL is available and restarts the service if not.

    Usage:

        self += HTTPWatchdogScript()
        self += HTTPServiceWatchdog(
            "application",
            script=self.path_to_script_that_starts_application,
            healthcheck_url="https://example.com",
            watchdog_script_path=self._.path
        )

    This creates a systemd service `application.service` that will be restarted
    if no HTTP request every 64s against :arg:`healthcheck_url` is successful.

    Arguments
    ---------

    * `service`: name of the systemd unit a watchdog is attached to.
    * `predefined_service`: whether the systemd unit is defined in another
      module, e.g. in NixOS itself (see below).
    * `script`: path to the script that gets started by the service.
    * `healthcheck_url`: the URL HTTP requests are issued against.
    * `healthcheck_timeout`: timeout for the HTTP request against :arg:`healthcheck_url`.
    * `check_interval`: time to sleep between two healthchecks.
    * `startup_check_interval`: time to sleep between healthchecks on startup.
    * `start_timeout`: how long until the first HTTP request must pass (see below for details).
    * `watchdog_interval`: interval between which one healthcheck must pass.

    Existing services
    -----------------

    If the watchdog needs to be added to an existing service, e.g.
    a service from NixOS itself, it can be done like this:

        self += HTTPServiceWatchdog(
            "docker-foobar",
            predefined_service=True,
            # ...
        )

    This only works if `systemd.services.docker-foobar.script` was declared
    by the module defining `docker-foobar.service`. If
    `systemd.services.docker-foobar.serviceConfig.ExecStart` is used by the module,
    this won't work.

    Inner workings
    --------------

    `systemd` provides a watchdog feature: the service needs to send `WATCHDOG=1` once in
    a given interval to `sd_notify(3)`. If that doesn't happen, the service will be aborted.
    Via an auto-restart defined in the unit, it will be started up again.

    This mechanism is used by this component:

    * It overrides the systemd service with a script that forks:
        * The parent process is the running application. This one
          must be started by the executable file passed to this
          component via :arg:`script`.
        * The child is a script that regularly checks if `healthcheck_url`
          is available and sends `WATCHDOG=1` if that's the case.

    * Certain applications can take a while until these are started up.
      When the unit is started, the forked off child process sends a request
      to :arg:`healthcheck_url` every :arg:`startup_check_interval` seconds.

      This is done until either

      * One request succeeds. The application is considered "up" by then
        and the watchdog starts.

        Before that's the case, `systemd` will list the unit as "activating"
        rather than "active".

      * A timeout set by :arg:`start_timeout` is reached. `systemd` will abort
        the application startup and terminate the unit. Because the unit is configured
        to restart, a startup will be attempted again.

      * If the startup attempts should be aborted eventually, this can be implemented
        via the options `StartLimitBurst`/`StartLimitIntervalSec` in the systemd unit.

        The options are documented in
        `systemd.unit(5) <https://www.freedesktop.org/software/systemd/man/latest/systemd.unit.html#StartLimitIntervalSec=interval>`.

        See below on how to modify the unit created by this component.

    * As long as :arg:`healthcheck_url` is up, a request will be issued every
      :arg:`check_interval` seconds and after that `WATCHDOG=1` will be sent
      to `sd_notify(3)`.

      As soon as the URL isn't successful anymore, an exponential backoff will
      be made: this means that after the request against :arg:`healthcheck_url`
      failed `n` times, 2^n seconds of sleep between two healthcheck checks are added
      to the :arg:`check_interval` seconds of sleep.

      For instance, after one failed request, the watchdog will wait 2^1=2 seconds
      in addition to :arg:`check_interval`.

      If :arg:`watchdog_interval` is 32 and :arg:`check_interval` is 1, there are
      three attempts to recover.

      I.e. 2^n seconds of sleep + 1s of sleep interval.
      For each of the three attempts so far this is `2^1 + 1 + 2^2 + 1 * 2^3 + 1 = 17s`.
      The timeout of 1s for the HTTP request is negligible in this example
      and thus left out.

      Because this is the fourth failure, a sleep of 2^4 + 1 seconds (=17s) will
      be started . Since 17+17 > 32, the watchdog will kill the process before another
      attempt can be made.

    Customize the systemd unit
    --------------------------

    The systemd unit will be written as Nix code to the target machine.
    This means that overrides are possible via the Nix module system. For instance,
    `StartLimitBurst`/`StartLimitIntervalSec` can be added like this:

        self += HTTPServiceWatchdog("foobar", ...)
        self += File(
            "/etc/local/nixos/customize-foobar.nix",
            content="{ systemd.services.foobar.serviceConfig = { "
                + "StartLimitIntervalSec = ...; StartLimitBurst = ... }; }"
        )

    Limitations
    -----------

    * The service injects a custom Python that's used by the watchdog script. It's highly
      recommended to override the PATH in :arg:`script`, e.g. by sourcing `/etc/profile`.
    * The running application must not kill the watchdog.
    """

    namevar = "service"
    predefined_service = batou.component.Attribute("literal", default=False)
    watchdog_script_path = batou.component.Attribute(str)
    script = batou.component.Attribute(str, default=None)

    healthcheck_url = batou.component.Attribute(str)
    healthcheck_timeout = batou.component.Attribute(int, default=2)
    check_interval = batou.component.Attribute(int, default=2)
    startup_check_interval = batou.component.Attribute(int, default=4)
    start_timeout = batou.component.Attribute(int, default=64)
    watchdog_interval = batou.component.Attribute(int, default=64)

    def configure(self):
        if self.predefined_service and self.script is not None:
            raise ValueError(
                f"batou_ext.http.HTTPServiceWatchdog({self.service}): cannot set 'script' if service is predefined."
            )
        if not self.predefined_service and self.script is None:
            raise ValueError(
                f"batou_ext.http.HTTPServiceWatchdog({self.service}): must set script if predefined_service is False."
            )

        self += batou.lib.file.File(
            f"/etc/local/nixos/{self.service}-watchdog.nix",
            content=pkg_resources.resource_string(
                __name__, "resources/http-watchdog.nix"
            ),
        )


class HTTPWatchdogScript(batou.component.Component):
    """
    Writes the wrapper script for the HTTP watchdog into the service user's home.
    Only needed in conjunction with :class:`batou_ext.http.HTTPServiceWatchdog`.
    """

    def configure(self):
        self += batou.lib.file.File(
            "watchdog-wrapper.py",
            mode=0o755,
            content=pkg_resources.resource_string(
                __name__, "resources/watchdog-wrapper.py"
            ),
        )
        self.path = self._.path



import batou.component
import batou.lib.mysql


class MySQLGeneric(batou.component.Component):
    """
    Allows to create default database as used e.g. for typical LAMP-projects

    Usage:
    self += batou_ext.mysql.MySQLGeneric(
        'mydb',
        username='myuser',
        password='mypasswod',
        admin_password='myadminpassword',
        allow_from_hostname='127.0.0.1',
        provide_as='myapplicationdatabase')
    """

    _required_params_ = {
        "username": "scott",
    }
    namevar = "database"
    username = None
    password = None
    admin_password = None

    provide_as = batou.component.Attribute(str, default="mysql")

    port = batou.component.Attribute(int, default=3306)

    # Used for GRANT-string
    allow_from_hostname = batou.component.Attribute(str, default="localhost")

    def configure(self):
        self.provide(self.provide_as, self)

        self.address = batou.utils.Address(self.host.fqdn, self.port)

        self += batou.lib.mysql.Database(
            self.database, admin_password=self.admin_password
        )
        self += batou.lib.mysql.User(
            self.username,
            password=self.password,
            admin_password=self.admin_password,
            allow_from_hostname=self.allow_from_hostname,
        )
        self += batou.lib.mysql.Grant(
            self.database,
            user=self.username,
            admin_password=self.admin_password,
            allow_from_hostname=self.allow_from_hostname,
        )



"""S3 operations

Example for downloading a file::


    self += batou_ext.s3.S3(endpoint_url="https://my.s3.endpoint",
                            access_key_id="1234567890ABCDEF",
                            secret_access_key="very_secure!1!")
    self.s3 = self._
    self += batou_ext.s3.Download("my/file/from/bucket",
                                    bucketname="mybucket",
                                    target="download.zip",
                                    s3=self.s3)

To make a bucket available, to e.g. the application::

    self += batou_ext.s3.Bucket('downloads', s3=self.s3)

"""
import os
from unittest.mock import Mock

import batou.component
import batou.utils
import boto3
import botocore.exceptions

from batou_ext.file import SymlinkAndCleanup


class S3(batou.component.Component):
    """Configuration for an S3 connection and its credentials.

    Keyword arguments:
    access_key_id     -- The S3 access key ID
    secret_access_key -- The S3 secret access key
    endpoint_url      -- The S3 enpoint's URL
    """

    _required_params_ = {
        "access_key_id": "value",
        "secret_access_key": "value",
        "endpoint_url": "value",
    }
    endpoint_url = batou.component.Attribute(str)
    access_key_id = batou.component.Attribute(str)
    secret_access_key = batou.component.Attribute(str)

    def configure(self):
        self.client = boto3.resource(
            "s3",
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            endpoint_url=self.endpoint_url,
        )


class Bucket(batou.component.Component):
    """Make given S3 bucket available.

    Usage::

        self += batou_ext.s3.Bucket('downloads', s3=self.s3)

    """

    _required_params_ = {
        "s3": Mock(),
    }
    namevar = "bucketname"
    s3 = batou.component.Attribute()

    def configure(self):
        self.bucket = self.s3.client.Bucket(self.bucketname)

    def verify(self):
        try:
            if not self.bucket.creation_date:
                raise batou.UpdateNeeded()
        except botocore.exceptions.ClientError as error:
            self.log(f"Error while checking for bucket {self.bucketname}")
            raise error

    def update(self):
        self.bucket.create()


class Download(batou.component.Component):
    """
    Downloads a file from a given S3 instance.

    Usage::

        self += batou_ext.s3.Download("my/file/from/bucket"
                                        s3,
                                        bucketname="mybucket")
    """

    _required_params_ = {
        "bucketname": "bucket",
        "target": "target",
        "s3": Mock(),
    }
    namevar = "key"
    s3 = batou.component.Attribute(str)
    key = batou.component.Attribute(str)
    bucketname = batou.component.Attribute(str)
    target = batou.component.Attribute(str)

    # Optional checksum. If given it is used as "must download" indicator,
    # otherwise the the file's ETAG from S3 is used.
    checksum = None

    def configure(self):
        if self.checksum:
            self.checksum_function, self.checksum = self.checksum.split(":")
        else:
            self.etag_file = self.target + ".etag"

        self.obj = self.s3.client.Object(self.bucketname, self.key)
        self.target = self.map(self.target)

    def verify(self):
        if not os.path.exists(self.target):
            raise batou.UpdateNeeded()
        if self.checksum:
            if self.checksum != batou.utils.hash(
                self.target, self.checksum_function
            ):
                raise batou.UpdateNeeded()
        else:
            if not os.path.exists(self.etag_file):
                raise batou.UpdateNeeded()
            with open(self.etag_file) as f:
                current_etag = f.read()
                try:
                    remote_etag = self.obj.e_tag
                except botocore.exceptions.ClientError:
                    # This usually means a 404. We'll pass the error to update
                    # So the output actually becomes useful.
                    raise batou.UpdateNeeded()
                if current_etag != remote_etag:
                    raise batou.UpdateNeeded()

    def update(self):
        self.obj.download_file(self.target)
        if self.checksum:
            target_checksum = batou.utils.hash(
                self.target, self.checksum_function
            )
            assert (
                self.checksum == target_checksum
            ), """\
Checksum mismatch!
expected: %s
got: %s""" % (
                self.checksum,
                target_checksum,
            )
        with open(self.etag_file, "w") as f:
            f.write(self.obj.e_tag)
